==> buckshot-0.8.7.jar: buckshot/backend/redis.clj <==
(ns buckshot.backend.redis
  (:require [buckshot.backend :as backend]
            [clojure.tools.reader.edn :as edn]
            [taoensso.carmine :as redis]))

(def key-prefix "buckshot:")

(def queue-key-prefix "buckshot:queue:")

(defn- ->key [x]
  (str key-prefix (name x)))

(defn- ->queue-key [x]
  (str queue-key-prefix (name x)))

(defn- next-by-queue [queue conn now]
  (first (redis/wcar conn
           (redis/zrangebyscore (->queue-key queue) "-inf" now "LIMIT" 0 1))))

(defrecord RedisBackend [conn listener listener-agent]
  backend/IBackend
  (add-job! [_ {:keys [id queue start] :as job}]
    (when (= 1 (redis/wcar conn
                 (redis/hsetnx (->key :jobs) id job)))
      (redis/wcar conn
        (redis/zadd (->queue-key queue) start id))
      job))
  (del-job! [_ {:keys [id queue] :as job}]
    (let [script "local job = redis.call('hget', 'buckshot:jobs', _:id)
                  if (not job) then
                    return nil
                  else
                    redis.call('hdel', 'buckshot:jobs', _:id)
                    redis.call('zrem', _:qk, _:id)
                    redis.call('zrem', 'buckshot:processing', _:id)
                    return 'ok'
                  end"
          r (redis/wcar conn
              (redis/lua script {:qk (->queue-key queue)} {:id id}))]
      (= "ok" r)))
  (get-jobs [_ ids]
    (when (seq ids)
      (redis/wcar conn
        (apply redis/hmget (->key :jobs) ids))))
  (next-job [this queues]
    (let [now (backend/now this)
          id (some #(next-by-queue % conn now) queues)]
      (redis/wcar conn
        (redis/hget (->key :jobs) id))))
  (claim-job! [this {:keys [id queue] :as job} worker-id]
    (let [claimed (assoc job :claimed (backend/now this) :worker worker-id)
          script "local score = redis.call('zscore', _:qk, _:id)
                  if (not score) then
                    return nil
                  else
                    redis.call('zrem', _:qk, _:id)
                    redis.call('zadd', 'buckshot:processing', score, _:id)
                    return 'ok'
                  end"
          r (redis/wcar conn
              (redis/lua script {:qk (->queue-key queue)} {:id id}))]
      (when (= "ok" r)
        (redis/wcar conn
          (redis/hset (->key :jobs) id claimed))
        claimed)))
  (queues [_]
    (->> (redis/wcar conn
           (redis/keys (str queue-key-prefix "*")))
         (map #(-> % (subs (count queue-key-prefix)) keyword))))
  (enqueued-ids [_ queues]
    (into {} (for [q queues]
               [q (redis/wcar conn
                    (redis/zrange (->queue-key q) 0 -1))])))
  (processing-ids [_]
    (redis/wcar conn
      (redis/zrange (->key :processing) 0 -1)))
  (publish! [_ channel message]
    (redis/wcar conn
      (redis/publish (->key channel) (pr-str message))))
  (subscribe! [_ channel f]
    (let [k (->key channel)
          msg-handler (fn [[type _ msg]]
                        (when (= type "message")
                          (f (edn/read-string msg))))]
      (send listener-agent (fn [_]
                             (redis/with-open-listener listener
                               (redis/subscribe k))
                             (swap! (:state listener) assoc k msg-handler)
                             nil))
      true))
  (now [_]
    (let [[seconds micros] (->> (redis/wcar conn
                                  (redis/time))
                                (map #(Long. %)))]
      (+ (* seconds 1000) (quot micros 1000)))))

(defn make
  "Makes a Redis Buckshot backend. 'opts' may contain :pool and :spec keys,
  in the same format as Carmine connection maps. By default, :pool is specified
  with unlimited active and idle connections."
  [opts]
  (let [defaults {:pool {:max-active -1 :max-idle -1}}
        conn (merge-with merge defaults opts)
        listener (redis/with-new-pubsub-listener (:spec conn) {})]
    (->RedisBackend conn listener (agent nil))))

==> buckshot-0.8.8.jar: buckshot/backend/redis.clj <==
(ns buckshot.backend.redis
  (:require [buckshot.backend :as backend]
            [clojure.tools.reader.edn :as edn]
            [taoensso.carmine :as redis]))

(def key-prefix "buckshot:")

(def queue-key-prefix "buckshot:queue:")

(defn- ->key [x]
  (str key-prefix (name x)))

(defn- ->queue-key [x]
  (str queue-key-prefix (name x)))

(defn- next-by-queue [queue conn now]
  (first (redis/wcar conn
           (redis/zrangebyscore (->queue-key queue) "-inf" now "LIMIT" 0 1))))

(defrecord RedisBackend [conn listener listener-agent]
  backend/IBackend
  (add-job! [_ {:keys [id queue start] :as job}]
    (when (= 1 (redis/wcar conn
                 (redis/hsetnx (->key :jobs) id job)))
      (redis/wcar conn
        (redis/zadd (->queue-key queue) start id))
      job))
  (del-job! [_ {:keys [id queue] :as job}]
    (let [script "local job = redis.call('hget', 'buckshot:jobs', _:id)
                  if (not job) then
                    return nil
                  else
                    redis.call('hdel', 'buckshot:jobs', _:id)
                    redis.call('zrem', _:qk, _:id)
                    redis.call('zrem', 'buckshot:processing', _:id)
                    return 'ok'
                  end"
          r (redis/wcar conn
              (redis/lua script {:qk (->queue-key queue)} {:id id}))]
      (= "ok" r)))
  (get-jobs [_ ids]
    (when (seq ids)
      (redis/wcar conn
        (apply redis/hmget (->key :jobs) ids))))
  (next-job [this queues]
    (let [now (backend/now this)
          id (some #(next-by-queue % conn now) queues)]
      (redis/wcar conn
        (redis/hget (->key :jobs) id))))
  (claim-job! [this {:keys [id queue] :as job} worker-id]
    (let [job (assoc job :claimed (backend/now this) :worker worker-id)
          script "local score = redis.call('zscore', _:qk, _:id)
                  if (not score) then
                    return nil
                  else
                    redis.call('hset', 'buckshot:jobs', _:id, _:job)
                    redis.call('zadd', 'buckshot:processing', score, _:id)
                    redis.call('zrem', _:qk, _:id)
                    return 'ok'
                  end"
          r (redis/wcar conn
              (redis/lua script {:qk (->queue-key queue)} {:id id :job job}))]
      (when (= "ok" r)
        job)))
  (queues [_]
    (->> (redis/wcar conn
           (redis/keys (str queue-key-prefix "*")))
         (map #(-> % (subs (count queue-key-prefix)) keyword))))
  (enqueued-ids [_ queues]
    (into {} (for [q queues]
               [q (redis/wcar conn
                    (redis/zrange (->queue-key q) 0 -1))])))
  (processing-ids [_]
    (redis/wcar conn
      (redis/zrange (->key :processing) 0 -1)))
  (publish! [_ channel message]
    (redis/wcar conn
      (redis/publish (->key channel) (pr-str message))))
  (subscribe! [_ channel f]
    (let [k (->key channel)
          msg-handler (fn [[type _ msg]]
                        (when (= type "message")
                          (f (edn/read-string msg))))]
      (send listener-agent (fn [_]
                             (redis/with-open-listener listener
                               (redis/subscribe k))
                             (swap! (:state listener) assoc k msg-handler)
                             nil))
      true))
  (now [_]
    (let [[seconds micros] (->> (redis/wcar conn
                                  (redis/time))
                                (map #(Long. %)))]
      (+ (* seconds 1000) (quot micros 1000)))))

(defn make
  "Makes a Redis Buckshot backend. 'opts' may contain :pool and :spec keys,
  in the same format as Carmine connection maps. By default, :pool is specified
  with unlimited active and idle connections."
  [opts]
  (let [defaults {:pool {:max-active -1 :max-idle -1}}
        conn (merge-with merge defaults opts)
        listener (redis/with-new-pubsub-listener (:spec conn) {})]
    (->RedisBackend conn listener (agent nil))))
